@@ -2,15 +2,25 @@ import json
 import time
 import logging
 import requests
-
+import importlib.util
+from typing import Union
 from .robot_state import RobotState
 from .eva_errors import eva_error, EvaError, EvaAutoRenewError
+from .version import __version__, sdk_is_compatible_with_robot
+
+# TODO: too big, install it manually if you want it
+has_pyt3d = importlib.util.find_spec('pytransform3d')
+if has_pyt3d:
+    import pytransform3d.rotations as pyrot  # type: ignore
 
 # TODO add more granular logs using __logger
 # TODO lots of sleeps in control_* de to the robot state being updated slowly after starting an action, can this be improved?
+# TODO technically, the default request timeout of the API is 60s, should we update `request_timeout`?
 
 
 class EvaHTTPClient:
+    __N_DIGITS = 8
+
     """
     Eva HTTP client
 
@@ -39,6 +49,7 @@ class EvaHTTPClient:
 
 
     def api_call_with_auth(self, *args, **kwargs):
+
         r = self.__api_request(*args, **kwargs)
 
         # Try creating a new session if we get an auth error and retrying the failed request
@@ -56,20 +67,36 @@ class EvaHTTPClient:
 
         return r
 
+    def api_call_no_auth(self, method, path, payload=None, **kwargs):
+        return self.__api_request(method, path, payload=payload, add_auth=False, **kwargs)
 
-    def __api_request(self, method, path, payload=None, headers=None, timeout=None):
-        if not headers:
-            headers = {'Authorization': 'Bearer {}'.format(self.session_token)}
+    def __api_request(self, method, path, payload=None, headers={}, timeout=None, version='v1', add_auth=True):
+        if headers:
+            headers = headers.copy()
+
+        if add_auth:
+            headers['Authorization'] = f'Bearer {self.session_token}'
+        headers['User-Agent'] = f'Automata EvaSDK/{__version__} (Python)'
+
+        api_path = path
+        if version is not None:
+            api_path = f'{version}/{path}'
 
         return requests.request(
-            method, 'http://{}/api/v1/{}'.format(self.host_ip, path),
+            method, 'http://{}/api/{}'.format(self.host_ip, api_path),
             data=payload, headers=headers,
             timeout=(timeout or self.request_timeout),
         )
 
     # API VERSIONS
     def api_versions(self):
-        r = self.__api_request('GET', 'versions')
+        r = self.api_call_no_auth('GET', 'versions', version=None)
+        if r.status_code != 200:
+            eva_error('api_versions request error', r)
+        return r.json()
+
+    def name(self):
+        r = self.api_call_no_auth('GET', 'name')
         if r.status_code != 200:
             eva_error('api_versions request error', r)
         return r.json()
@@ -93,8 +120,12 @@ class EvaHTTPClient:
 
     def auth_create_session(self):
         self.__logger.debug('Creating session token')
-        # Bypass api_call_with_auth to avoid getting in a 401 loop
-        r = self.__api_request('POST', 'auth', payload=json.dumps({'token': self.api_token}), headers={})
+
+        err = self._check_version_compatibility()
+        if err is not None:
+            self.__logger.error(f'incompatible SDK: {err}')
+
+        r = self.api_call_no_auth('POST', 'auth', payload=json.dumps({'token': self.api_token}))
 
         if r.status_code != 200:
             eva_error('auth_create_session request error', r)
@@ -107,6 +138,7 @@ class EvaHTTPClient:
 
     def auth_invalidate_session(self):
         self.__logger.debug('Invalidating session token {}'.format(self.session_token))
+        # Bypass api_call_with_auth to avoid getting in a 401 loop
         r = self.__api_request('DELETE', 'auth')
 
         if r.status_code != 204:
@@ -147,7 +179,7 @@ class EvaHTTPClient:
     def config_update(self, update):
         r = self.api_call_with_auth(
             'POST', 'config/update', update,
-            headers={'Content-Type': 'application/x.automata-update'}, timeout=30
+            headers={'Content-Type': 'application/x.automata-update'}, timeout=120
         )
         if r.status_code != 200:
             eva_error('config_update error', r)
@@ -156,7 +188,7 @@ class EvaHTTPClient:
 
     # GPIO
     def gpio_set(self, pin, status):
-        r = self.__globals_editing(keys='outputs.' + pin, values=status)
+        r = self._globals_edit(keys='outputs.' + pin, values=status)
         if r.status_code != 200:
             eva_error('gpio_set error', r)
 
@@ -169,16 +201,21 @@ class EvaHTTPClient:
         return self.data_snapshot_property('global.{}s'.format(pin_type))[pin]
 
 
-    # GPIO helper function
-    def __globals_editing(self, keys, values):
+    def _globals_edit(self, keys, values):
         data = {'changes': []}
         if (isinstance(keys, list) and isinstance(values, list)):
-            [data['changes'].append({'key': c[0], 'value': c[1]}) for c in zip(keys, values)]
+            data['changes'] = [{'key': k, 'value': v} for k, v in zip(keys, values)]
         else:
             data['changes'].append({'key': keys, 'value': values})
         data = json.dumps(data)
-        r = self.api_call_with_auth('POST', 'data/globals', data)
-        return r
+        return self.api_call_with_auth('POST', 'data/globals', data)
+
+
+    def globals_edit(self, keys, values):
+        r = self._globals_edit(keys, values)
+        if r.status_code != 200:
+            eva_error('globals_edit error', r)
+        return r.json()
 
 
     # TOOLPATHS
@@ -222,13 +259,13 @@ class EvaHTTPClient:
 
 
     def toolpaths_use_saved(self, toolpathId):
-        r = self.api_call_with_auth('POST', 'toolpaths/{}/use'.format(toolpathId))
+        r = self.api_call_with_auth('POST', 'toolpaths/{}/use'.format(toolpathId), timeout=300)
         if r.status_code != 200:
             eva_error('toolpaths_use_saved error', r)
 
 
     def toolpaths_use(self, toolpathRepr):
-        r = self.api_call_with_auth('POST', 'toolpath/use', json.dumps({'toolpath': toolpathRepr}))
+        r = self.api_call_with_auth('POST', 'toolpath/use', json.dumps({'toolpath': toolpathRepr}), timeout=300)
         if r.status_code != 200:
             eva_error('toolpaths_use error', r)
 
@@ -298,6 +335,8 @@ class EvaHTTPClient:
 
             if robot_state == RobotState.ERROR:
                 eva_error('Eva is in error control state')
+            if robot_state == RobotState.COLLISION:
+                eva_error('Eva has encountered a collision. Please acknowledge collision.')
             elif robot_state == parsed_goal:
                 return
 
@@ -328,18 +367,19 @@ class EvaHTTPClient:
             time.sleep(0.1)     # sleep for small period to avoid race condition between updating cache and reading state
             self.control_wait_for(RobotState.READY)
 
-    def control_go_to(self, joints, wait_for_ready=True, velocity=None, duration=None, mode='teach'):
+
+    def control_go_to(self, joints, wait_for_ready=True, max_speed=None, time_sec=None, mode='teach'):
         body = {'joints': joints, 'mode': mode}
-        if velocity is not None:
-            body['velocity'] = velocity
-        elif duration is not None:
-            body['time'] = duration
+        if max_speed is not None:
+            body['max_speed'] = max_speed
+        if time is not None:
+            body['time'] = time_sec
 
         r = self.api_call_with_auth('POST', 'controls/go_to', json.dumps(body))
         if r.status_code != 200:
             eva_error('control_go_to error', r)
         elif wait_for_ready:
-            time.sleep(0.1)     # sleep for small period to avoid race condition between updating cache and reading state
+            time.sleep(0.1)     # sleep for small period to avoid race condition between cache updating and reading state
             self.control_wait_for(RobotState.READY)
 
 
@@ -352,13 +392,13 @@ class EvaHTTPClient:
             self.control_wait_for(RobotState.PAUSED)
 
 
-    def control_resume(self, wait_for_ready=True):
+    def control_resume(self, wait_for_running=True):
         r = self.api_call_with_auth('POST', 'controls/resume')
         if r.status_code != 200:
             eva_error('control_resume error', r)
-        elif wait_for_ready:
+        elif wait_for_running:
             time.sleep(0.1)     # sleep for small period to avoid race condition between updating cache and reading state
-            self.control_wait_for(RobotState.READY)
+            self.control_wait_for(RobotState.RUNNING)
 
 
     def control_cancel(self, wait_for_ready=True):
@@ -387,9 +427,42 @@ class EvaHTTPClient:
             time.sleep(0.1)     # sleep for small period to avoid race condition between updating cache and reading state
             self.control_wait_for(RobotState.READY)
 
+    # COLLISION DETECTION
+    def control_configure_collision_detection(self, enabled, sensitivity):
+        """
+        Allows toggling on/off and setting the sensitivity of collision detection
+        """
+        if sensitivity not in ['low', 'medium', 'high']:
+            raise ValueError('collision_detection error, no such sensitivity ' + sensitivity)
+        if enabled not in [True, False]:
+            raise ValueError('collision_detection error, must be True/False')
+        r = self.api_call_with_auth('POST', 'controls/collision_detection',
+                                    json.dumps({'enabled': enabled, 'sensitivity': sensitivity}))
+        if r.status_code != 204:
+            eva_error('control_collision_detection error', r)
+
+    def control_acknowledge_collision(self, wait_for_ready=True):
+        """
+        When a collision is encountered, it must be acknowledged before the robot can be reset
+        """
+        r = self.api_call_with_auth('POST', 'controls/acknowledge_collision')
+        if r.status_code != 204:
+            eva_error('control_acknowledge_collision error', r)
+        elif wait_for_ready:
+            time.sleep(0.1)     # sleep for small period to avoid race condition between updating cache and reading state
+            self.control_wait_for(RobotState.READY)
 
     # CALCULATIONS
-    def calc_forward_kinematics(self, joints, fk_type='both', tcp_config=None):
+    def __check_calculation(self, r, name, kind):
+        res = r.json()
+        if res[kind]['result'] != 'success':
+            eva_error(f'{name} error: {res[kind]["error"]}')
+        return res[kind]
+
+    def calc_forward_kinematics(self, joints, fk_type=None, tcp_config=None):
+        if fk_type is not None:
+            self.__logger.warn('deprecated fk_type keyword, now all FK data is being returned')
+
         body = {'joints': joints}
         if tcp_config is not None:
             body['tcp_config'] = tcp_config
@@ -398,17 +471,60 @@ class EvaHTTPClient:
 
         if r.status_code != 200:
             eva_error('calc_forward_kinematics error', r)
+        return self.__check_calculation(r, 'calc_forward_kinematics', 'fk')
 
-        if (fk_type == 'position') or (fk_type == 'orientation'):
-            return r.json()['fk'][fk_type]
-        elif (fk_type == 'both'):
-            return r.json()['fk']
+    def __ensure_pyt3d(self):
+        if not has_pyt3d:
+            raise ImportError('''this feature is provided by `pytransform3d`, which we do not ship by default anymore due to its dependencies
+
+If you require this feature, just install it yourself (e.g. `pip3 install pytransform3d`, add it to your `Pipfile`, etc)
+''')
+
+    def __normalize(self, vec):
+        if has_pyt3d:
+            return [round(num, self.__N_DIGITS) for num in pyrot.check_quaternion(vec)]
+        norm = sum(vec)
+        if norm == 0:
+            norm = 1
+        return [round(n / norm, self.__N_DIGITS) for n in vec]
+
+    def calc_inverse_kinematics(self, guess, target_position, target_orientation, tcp_config=None,
+                                orientation_type=None):
+        """
+        End-effector orientation (target_orientation) can be provided in several standard formats,
+        by specifying the orinetation_type (default is None):
+        - 'matrix': rotation matrix -> 3x3 array, in row major order
+        - 'axis_angle': axis angle -> {'angle': float, 'x': float, 'y': float, 'z': float}
+        - 'euler_zyx': {yaw, pitch, roll} Euler (Tait-Bryan) angles -> {'yaw': float, 'pitch': float, 'roll': float}
+        - 'quat': quaternion -> {'w': float, 'x': float, 'y': float, 'z': float}
+        - None: defaults to quaternion
+        Conversion relies on pytransform3d library
+        """
+        quat_not_normed = None
+
+        if orientation_type == 'matrix':
+            self.__ensure_pyt3d()
+            quat_not_normed = pyrot.quaternion_from_matrix(pyrot.check_matrix(target_orientation))
+        elif orientation_type == 'axis_angle':
+            self.__ensure_pyt3d()
+            axis_angle = [target_orientation['x'], target_orientation['y'], target_orientation['z'],
+                          target_orientation['angle']]
+            quat_not_normed = pyrot.quaternion_from_axis_angle(pyrot.check_axis_angle(axis_angle))
+        elif orientation_type == 'euler_zyx':
+            self.__ensure_pyt3d()
+            euler_zyx = [target_orientation['yaw'], target_orientation['pitch'], target_orientation['roll']]
+            matrix = pyrot.matrix_from_euler_zyx(euler_zyx)
+            quat_not_normed = pyrot.quaternion_from_matrix(pyrot.check_matrix(matrix))
+        elif orientation_type == 'quat' or orientation_type is None:
+            quat_not_normed = [target_orientation['w'], target_orientation['x'], target_orientation['y'],
+                               target_orientation['z']]
         else:
-            eva_error('calc_forward_kinematics invalid fk_type {}'.format(fk_type), r)
+            eva_error(f'calc_inverse_kinematics invalid "{orientation_type}" orientation_type')
 
+        quat_normed = self.__normalize(quat_not_normed)
+        quaternion = {'w': quat_normed[0], 'x': quat_normed[1], 'y': quat_normed[2], 'z': quat_normed[3]}
 
-    def calc_inverse_kinematics(self, guess, target_position, target_orientation, tcp_config=None):
-        body = {'guess': guess, 'position': target_position, 'orientation': target_orientation}
+        body = {'guess': guess, 'position': target_position, 'orientation': quaternion}
         if tcp_config is not None:
             body['tcp_config'] = tcp_config
 
@@ -416,8 +532,7 @@ class EvaHTTPClient:
 
         if r.status_code != 200:
             eva_error('inverse_kinematics error', r)
-        return r.json()
-
+        return self.__check_calculation(r, 'calc_inverse_kinematics', 'ik')['joints']
 
     def calc_nudge(self, joints, direction, offset, tcp_config=None):
         body = {'joints': joints, 'direction': direction, 'offset': offset}
@@ -428,8 +543,7 @@ class EvaHTTPClient:
 
         if r.status_code != 200:
             eva_error('calc_nudge error', r)
-        return r.json()['nudge']['joints']
-
+        return self.__check_calculation(r, 'calc_nudge', 'nudge')['joints']
 
     def calc_pose_valid(self, joints, tcp_config=None):
         body = {'joints': joints}
@@ -452,4 +566,19 @@ class EvaHTTPClient:
 
         if r.status_code != 200:
             eva_error('calc_rotate error', r)
-        return r.json()
+        return self.__check_calculation(r, 'calc_rotate', 'rotate')['joints']
+
+
+    def _check_version_compatibility(self) -> Union[str, None]:
+        """Checks the current version of the application against that of Eva's software version.
+        Returns:
+            Union[str, None]: error message, None if versions are compatible.
+        """
+        version_r = self.api_call_no_auth('GET', 'versions', version=None)
+        if version_r.status_code != 200:
+            return 'request error when fetching versions'
+        robot_version = version_r.json()['robot']
+        err = sdk_is_compatible_with_robot(robot_version)
+        if err is not None:
+            return err
+        return None
